"""Module containing the main base classes"""
from typing import Optional, Union, Any, Dict, List

import redis
from pydantic import BaseModel

from pydantic_redis.config import RedisConfig
from pydantic_redis.utils import bytes_to_string


class _AbstractStore(BaseModel):
    """
    An abstract class of a store
    """
    name: str
    redis_config: RedisConfig
    redis_store: Optional[redis.Redis] = None
    life_span_in_seconds: Optional[int] = None

    class Config:
        arbitrary_types_allowed = True
        orm_mode = True


class _AbstractModel(BaseModel):
    """
    An abstract class to help with typings for Model class
    """
    _store: _AbstractStore
    _primary_key_field: str

    @staticmethod
    def serialize_partially(data: Optional[Dict[str, Any]]):
        """Converts non primitive data types into str"""
        return {key: (value if isinstance(value, (str, float, int)) and not isinstance(value, (bool,)) else str(value))
                for key, value in data.items()}

    @staticmethod
    def deserialize_partially(data: Optional[Dict[bytes, Any]]):
        """Converts non primitive data types into str"""
        return {bytes_to_string(key): (bytes_to_string(value) if isinstance(value, (bytes,)) else value)
                for key, value in data.items()}

    @classmethod
    def get_primary_key_field(cls):
        """Gets the protected _primary_key_field"""
        return cls._primary_key_field

    @classmethod
    def insert(cls, data: Union[List[Any], Any]):
        raise NotImplementedError("insert should be implemented")

    @classmethod
    def update(cls, primary_key_value: Union[Any, Dict[str, Any]], data: Dict[str, Any]):
        raise NotImplementedError("update should be implemented")

    @classmethod
    def delete(cls, primary_key_value: Union[Any, Dict[str, Any]]):
        raise NotImplementedError("delete should be implemented")

    @classmethod
    def select(cls, columns: Optional[List[str]] = None):
        """Should later allow AND, OR"""
        raise NotImplementedError("select should be implemented")

    class Config:
        arbitrary_types_allowed = True



"""Module containing the main config classes"""

from pydantic import BaseModel


class RedisConfig(BaseModel):
    host: str = 'localhost'
    port: int = 6379
    db: int = 0

    class Config:
        orm_mode = True



"""Module containing the model classes"""
import uuid
from typing import Optional, List, Any, Union, Dict

from pydantic_redis.abstract import _AbstractModel
from pydantic_redis.utils import bytes_to_string


class Model(_AbstractModel):
    """
    The section in the store that saves rows of the same kind
    """

    @classmethod
    def __get_primary_key(cls, primary_key_value: Any):
        """
        Returns the primary key value concatenated to the table name for uniqueness
        """
        table_name = cls.__name__.lower()
        return f"{table_name}_%&_{primary_key_value}"

    @classmethod
    def get_table_index_key(cls):
        """Returns the key in which the primary keys of the given table have been saved"""
        table_name = cls.__name__.lower()
        return f"{table_name}__index"

    @classmethod
    def insert(cls, data: Union[List[_AbstractModel], _AbstractModel], life_span_seconds: Optional[float] = None):
        """
        Inserts a given row or sets of rows into the table
        """
        life_span = life_span_seconds if life_span_seconds is not None else cls._store.life_span_in_seconds
        pipeline = cls._store.redis_store.pipeline()
        data_list = []

        if isinstance(data, list):
            data_list = data
        elif isinstance(data, _AbstractModel):
            data_list = [data]

        for record in data_list:
            primary_key_value = getattr(record, cls._primary_key_field, str(uuid.uuid4()))
            name = cls.__get_primary_key(primary_key_value=primary_key_value)
            mapping = cls.serialize_partially(record.dict())
            pipeline.hset(name=name, mapping=mapping)
            pipeline.expire(name=name, time=life_span)
            # save the primary key in an index
            table_index_key = cls.get_table_index_key()
            pipeline.sadd(table_index_key, name)
            pipeline.expire(table_index_key, time=life_span)

        return pipeline.execute()

    @classmethod
    def update(cls, _id: Any, data: Dict[str, Any],
               life_span_seconds: Optional[float] = None):
        """
        Updates a given row or sets of rows in the table
        """
        life_span = life_span_seconds if life_span_seconds is not None else cls._store.life_span_in_seconds
        pipeline = cls._store.redis_store.pipeline()

        if isinstance(data, dict):
            name = cls.__get_primary_key(primary_key_value=_id)
            pipeline.hset(name=name, mapping=cls.serialize_partially(data))
            pipeline.expire(name=name, time=life_span)
            # save the primary key in an index
            table_index_key = cls.get_table_index_key()
            pipeline.sadd(table_index_key, name)
            pipeline.expire(table_index_key, time=life_span)

        return pipeline.execute()

    @classmethod
    def delete(cls, ids: Union[Any, List[Any]]):
        """
        deletes a given row or sets of rows in the table
        """
        pipeline = cls._store.redis_store.pipeline()
        primary_keys = []

        if isinstance(ids, list):
            primary_keys = ids
        elif ids is not None:
            primary_keys = [ids]

        names = [cls.__get_primary_key(primary_key_value=primary_key_value) for primary_key_value in primary_keys]
        pipeline.delete(*names)
        # remove the primary keys from the index
        table_index_key = cls.get_table_index_key()
        pipeline.srem(table_index_key, *names)
        return pipeline.execute()

    @classmethod
    def select(cls, columns: Optional[List[str]] = None, ids: Optional[List[Any]] = None):
        """
        Selects given rows or sets of rows in the table
        """
        pipeline = cls._store.redis_store.pipeline()
        keys = ()

        if ids is None:
            # get all keys in the table immediately so don't use a pipeline
            table_index_key = cls.get_table_index_key()
            keys = cls._store.redis_store.sscan_iter(name=table_index_key)
        else:
            keys = (cls.__get_primary_key(primary_key_value=primary_key) for primary_key in ids)

        for key in keys:
            if columns is None:
                pipeline.hgetall(name=key)
            else:
                pipeline.hmget(name=key, keys=columns)
        response = pipeline.execute()

        if len(response) == 0:
            return None

        if isinstance(response, list) and columns is None:
            return [cls(**cls.deserialize_partially(record)) for record in response]
        elif isinstance(response, list) and columns is not None:
            return [{field: bytes_to_string(record[index]) for index, field in enumerate(columns)} for record in
                    response]
        return cls(**cls.deserialize_partially(response)) if isinstance(response, dict) else response



"""Module containing the store classes"""
from typing import Dict, Optional, Any

import redis

from pydantic_redis.abstract import _AbstractStore
from pydantic_redis.config import RedisConfig
from pydantic_redis.model import Model


class Store(_AbstractStore):
    """
    A store that allows a declarative way of querying for data in redis
    """
    models: Dict[str, type(Model)] = {}

    def __init__(self, name: str, redis_config: RedisConfig, redis_store: Optional[redis.Redis] = None,
                 life_span_in_seconds: Optional[int] = None, **data: Any):
        super().__init__(name=name, redis_config=redis_config, redis_store=redis_store,
                         life_span_in_seconds=life_span_in_seconds, **data)
        self.redis_store = redis.Redis(**self.redis_config.dict())

    def register_model(self, model_class: type(Model)):
        """Registers the model to this store"""
        if not isinstance(model_class.get_primary_key_field(), str):
            raise NotImplementedError(f"{model_class.__name__} should have a _primary_key_field")

        model_class._store = self
        self.models[model_class.__name__.lower()] = model_class

    def model(self, name: str) -> Model:
        """Gets a model by name: case insensitive"""
        return self.models[name.lower()]



"""Tests for the redis orm"""

import unittest
from datetime import date

from pydantic_redis.config import RedisConfig
from pydantic_redis.model import Model
from pydantic_redis.store import Store


class Book(Model):
    _primary_key_field: str = 'title'
    title: str
    author: str
    published_on: date
    in_stock: bool = True


books = [
    Book(title="Oliver Twist", author='Charles Dickens', published_on=date(year=1215, month=4, day=4),
         in_stock=False),
    Book(title="Great Expectations", author='Charles Dickens', published_on=date(year=1220, month=4, day=4)),
    Book(title="Jane Eyre", author='Charles Dickens', published_on=date(year=1225, month=6, day=4), in_stock=False),
    Book(title="Wuthering Heights", author='Jane Austen', published_on=date(year=1600, month=4, day=4)),
]


class ModelWithoutPrimaryKey(Model):
    title: str


class TestRedisOrm(unittest.TestCase):
    """
    Tests for Redis Orm
    """

    @classmethod
    def setUpClass(cls) -> None:
        cls.store = Store(name='sample', redis_config=RedisConfig(db=5), life_span_in_seconds=3600)
        cls.store.register_model(Book)

    def tearDown(self) -> None:
        """Tear down a few things"""
        keys = [f"book_%&_{book.title}" for book in books]
        self.store.redis_store.delete(*keys)

    def test_register_model_without_primary_key(self):
        """Throws error when a model without the _primary_key_field class variable set is registered"""
        self.assertRaisesRegex(AttributeError, '_primary_key_field', self.store.register_model,
                               ModelWithoutPrimaryKey)
        ModelWithoutPrimaryKey._primary_key_field = None
        self.assertRaisesRegex(Exception, 'should have a _primary_key_field', self.store.register_model,
                               ModelWithoutPrimaryKey)

    def test_bulk_insert(self):
        """Providing a list of Model instances to the insert method inserts the records in redis"""
        keys = [f"book_%&_{book.title}" for book in books]
        self.store.redis_store.delete(*keys)

        for key in keys:
            book_in_redis = self.store.redis_store.hgetall(name=key)
            self.assertEqual(book_in_redis, {})

        Book.insert(books)

        pipeline = self.store.redis_store.pipeline()
        for key in keys:
            pipeline.hgetall(name=key)
        books_in_redis = pipeline.execute()
        books_in_redis_as_models = [Book(**Book.deserialize_partially(book)) for book in books_in_redis]
        self.assertEqual(books, books_in_redis_as_models)

    def test_insert_single(self):
        """
        Providing a single Model instance
        """
        key = f"book_%&_{books[0].title}"
        book = self.store.redis_store.hgetall(name=key)
        self.assertEqual(book, {})

        Book.insert(books[0])

        book = self.store.redis_store.hgetall(name=key)
        book_as_model = Book(**Book.deserialize_partially(book))
        self.assertEqual(books[0], book_as_model)

    def test_select_default(self):
        """Selecting without arguments returns all the book models"""
        Book.insert(books)
        response = Book.select()
        sorted_books = sorted(books, key=lambda x: x.title)
        sorted_response = sorted(response, key=lambda x: x.title)
        self.assertEqual(sorted_books, sorted_response)

    def test_select_some_columns(self):
        """
        Selecting some columns returns a list of dictionaries of all books models with only those columns
        """
        Book.insert(books)
        books_dict = {book.title: book for book in books}
        columns = ['title', 'author', 'in_stock']
        response = Book.select(columns=['title', 'author', 'in_stock'])
        response_dict = {book['title']: book for book in response}

        for title, book in books_dict.items():
            book_in_response = response_dict[title]
            self.assertIsInstance(book_in_response, dict)
            self.assertEqual(sorted(book_in_response.keys()), sorted(columns))
            for column in columns:
                self.assertEqual(f"{book_in_response[column]}", f"{getattr(book, column)}")

    def test_select_some_ids(self):
        """
        Selecting some ids returns only those elements with the given ids
        """
        Book.insert(books)
        ids = [book.title for book in books[:2]]
        response = Book.select(ids=ids)
        self.assertEqual(response, books[:2])

    def test_update(self):
        """
        Updating an item of a given primary key updates it in redis
        """
        Book.insert(books)
        title = books[0].title
        new_author = 'John Doe'
        key = f"book_%&_{title}"
        old_book_data = self.store.redis_store.hgetall(name=key)
        old_book = Book(**Book.deserialize_partially(old_book_data))
        self.assertEqual(old_book, books[0])
        self.assertNotEqual(old_book.author, new_author)

        Book.update(_id=title, data={"author": "John Doe"})

        book_data = self.store.redis_store.hgetall(name=key)
        book = Book(**Book.deserialize_partially(book_data))
        self.assertEqual(book.author, new_author)
        self.assertEqual(book.title, old_book.title)
        self.assertEqual(book.in_stock, old_book.in_stock)
        self.assertEqual(book.published_on, old_book.published_on)

    def test_delete_multiple(self):
        """
        Providing a list of ids to the delete function will remove the items from redis
        """
        Book.insert(books)
        books_to_delete = books[:2]
        books_left_in_db = books[2:]

        ids_to_delete = [book.title for book in books_to_delete]
        ids_to_leave_intact = [book.title for book in books_left_in_db]

        keys_to_delete = [f"book_%&_{_id}" for _id in ids_to_delete]
        keys_to_leave_intact = [f"book_%&_{_id}" for _id in ids_to_leave_intact]

        Book.delete(ids=ids_to_delete)

        for key in keys_to_delete:
            deleted_book_in_redis = self.store.redis_store.hgetall(name=key)
            self.assertEqual(deleted_book_in_redis, {})

        pipeline = self.store.redis_store.pipeline()
        for key in keys_to_leave_intact:
            pipeline.hgetall(name=key)
        books_in_redis = pipeline.execute()
        books_in_redis_as_models = [Book(**Book.deserialize_partially(book)) for book in books_in_redis]
        self.assertEqual(books_left_in_db, books_in_redis_as_models)


if __name__ == '__main__':
    unittest.main()
